//-----------------------------------------------------------------------------
// File: Assault.h
//
// Desc: Island assault script
//
// Copyright (c) 2000-2008 Black Element Software. All rights reserved
//-----------------------------------------------------------------------------

/** \addtogroup StartegicAPI */
/**@{*/
//! Units from AIs carriers
class AssaultUnit extends CustomSerializable
{
	int AssemblyID;
	bool Deployed;

	//! Invalidate AssaultUnit
	void Invalidate()
	{
		AssemblyID = -1;
		Deployed = false;
	}

	//! Auto Equip Assault Unit
	void AutoEquip( bool manta, CCPlayer player, int assaultType )
	{
		// first remove all items
		EntityBasic assembly = player.GetAssembly( AssemblyID );

		int equipCnt = assembly.GetEquipmentCount();

		if ( equipCnt != 0 )
		{
			for ( int i = 0; i < equipCnt; i++ )
			{
				int item = assembly.GetEquipment( i );
				player.GetCarrierStorage().AddItem( EGAMESTORAGE_OBJECTIVE, item, 1 );
			}
		}

		assembly.SafeClearEquipment();

		if ( GetCampaign() )
			GetCampaign().EquipUnitGeneric( assembly, manta, player, assaultType );

		// enable all equipped items
		if ( Deployed && assembly.GetEntity())
			assembly.GetEntity().EnableItems();
	}

	/*! BasicUnit */
	/* \return EntityBasic */
	EntityBasic BasicUnit(CCPlayer player)
	{
		EntityBasic assembly = player.GetAssembly( AssemblyID );
		if ( assembly )
			return assembly;

		return NULL;
	}

	Unit RelUnit(CCPlayer player)
	{
		//EntityBasic assembly = player.GetAssembly( AssemblyID );
		EntityBasic assembly = BasicUnit( player );

		if ( assembly )
		{
			Unit runit = assembly.GetEntity();
			return runit;
		}
		return NULL;
	}
};

//! solve AIs assaults on islands
class Assault extends CustomSerializable
{
	const int BATTLE_ACTION = 0;
	const int BATTLE_DISTANT = 1;

	const int STATE_NONE = 0;
	const int STATE_ASSAULT = 1;
	const int STATE_DOCK = 2;

	const float MAX_FIGHT_DURATION = 6000.0;  

	const float GATHER_DIST_SQ = 1920.0;
	//const float GATHER_DIST_SQ = 409600.0;

	const float MAX_CC_HACKING_TIME = 100.0;

	IslandModel Island;
	CCPlayer MyPlayer;

	int Mode;
	int Progress;

	bool Panic;
	bool SwitchedToAction;
	bool WaitingForTeleport;
	bool HackingCC;

	int CurrTeleportIndex;
	int CurrTeleportStep;

	bool PostponeOneFrame;

	float Time;

	// distant fight over a point
	float FightTimeStamp;
	float FightDuration;
	float PtAAPower;
	float PtAADefense;
	float PtATPower;
	float PtATDefense;
	float PlayerAAPower;
	float PlayerATPower;
	float PlayerAADefense;
	float PlayerATDefense;
	float PlayerHealing;
	float CapMod;

	float CCTimeStamp;

	const int MAX_UNITS = 8;
	
	AssaultUnit Units[MAX_UNITS];

	IslandKeyPoint CurrentPoint;
	IslandKeyPoint FinalPoint;
	IslandKeyPoint VisitedPoints[32];
	int VisitedCnt;
	IslandKeyPoint Path[32];
	int PathLen;
	int PathPos;
	IslandKeyPoint Scramblers[8];
	int ScramblerCnt;

	bool TestMode;

	void Assault()
	{
		Progress = STATE_NONE;
		HackingCC = false;

		for( int i = 0; i < MAX_UNITS; i++ )
		{
			AssaultUnit unit = new AssaultUnit;
			unit.Invalidate();
			Units[i] = unit;
		}
	}

	void ~Assault()
	{
		for( int i = 0; i < MAX_UNITS; i++ )
		{
			AssaultUnit unit = Units[i];
			delete unit;
		}
	}

	//-------------------------------------------
	void ShowUnitPlanInDebug( int slot, string plan )
	{
		string temp;
		if ( slot <= 5)
			temp = "StrategyRoot.Manta"
		else
			temp = "StrategyRoot.Walrus"				

	
		temp += itoa(slot)+"Plan";
		widget unit = FindWidget( NULL, temp);

		widget unitPlan = FindWidget( NULL, temp );
		if ( unitPlan != NULL)
			SetWidgetText( unitPlan, 0, plan );
		else
		{
			Print( temp );
			Print( slot );
			Print( MAX_UNITS );
		}
	}



	//! switch assault mode to action
	void ProcessActionSwitch()
	{
		// kill units on visited points
		for( int i = 0; i < VisitedCnt; ++i )
		{
			IslandKeyPoint pt = VisitedPoints[i];
			string tmpLog = "Byly zniceny jednotky kolem keypointu: " + pt.GetName();
			SLog( tmpLog );
			pt.KillAllUnitsAround();
		}
/*
		// start deploy of all units that should be outside Carrier
		for( i = 0; i < MAX_UNITS; i++ )
		{
			AssaultUnit u = Units[i];
			if ( u.AssemblyID < 0 )
				continue;

			EntityBasic assembly = MyPlayer.GetAssembly( u.AssemblyID );
			if ( assembly.IsDeployable() )
			{
				// reequip (?)
				//assembly.ClearAllEquipment();
				MyPlayer.GetCarrier().AssemblyDeploy( assembly.GetSlot() );
			}
		}
*/
		// prepare teleport counter
		CurrTeleportIndex = 0;
		CurrTeleportStep = 1;
		WaitingForTeleport = true;
	}

	//! teleport unit to right keypoint after mode has been switched to action
	void TeleportAfterModeSwitch( AssaultUnit unit )
	{
		if ( !CurrentPoint || !unit.Deployed )
			return;

 		Unit runit = unit.RelUnit(MyPlayer);
 		if (runit == NULL)
 			return;

		float step = M_PI2 / 8.0;
		vector pos = ZeroVec;
		int iter = 0;
		while( pos == ZeroVec && iter < 16 )
		{
			float angle = step * CurrTeleportIndex;
			CurrTeleportIndex += 1;
			float dist = CurrTeleportStep * 480.0;
			if( CurrTeleportIndex >= 8 )
			{
				CurrTeleportStep += 1;
				CurrTeleportIndex = 0;
			}

			vector dir = Vector( 1, 0, 0 );
			dir = RotateVectorXY( dir, angle ) * 480.0;
			pos = CurrentPoint.GetPosition() + dir;
			pos = FindClosestReachable( pos, 960.0, runit.GetAI() );
			iter++;
		}
	
		if ( pos != ZeroVec )
		{
			//Teleport Units after deploying can be seen by players
			//runit.Teleport( pos, 0 );
			unit.BasicUnit(MyPlayer).InstantDeploy( pos, 0 );
			CommandCenter cc = g_Game.GetCommandCenter();
			if ( cc )
			{
				vector CCpos = cc.GetPosition();
				runit.GoTo( CCpos, REPLACE );

#ifdef DEVELOPER
				ShowUnitPlanInDebug( unit.BasicUnit( MyPlayer).GetSlot()+1, "CC");
				widget reason = FindWidget( NULL, "StrategyRoot.MA_KeypointNameValue" );
				SetWidgetText( reason , 0, "Je CC" );
#endif
			}
			else
			{
				runit.Dock( REPLACE );

#ifdef DEVELOPER
				ShowUnitPlanInDebug( unit.BasicUnit( MyPlayer).GetSlot()+1, "DOCK");
				reason = FindWidget( NULL, "StrategyRoot.MA_KeypointNameValue" );
				SetWidgetText( reason , 0, "Neni CC" );
#endif
			}
		}
		else
		{
			Print( "Failed to teleport assault unit to location" );
			Print( CurrentPoint.GetPosition() );
		}
	}

	void TowardsNextPoint()
	{
		// select next point
		Print("TowardsNextPoint");
		IslandKeyPoint final = FinalPoint;
		if ( ScramblerCnt > 0 )
		{
			//TEST: all scramblers, one by one
			final = Scramblers[0];
			ScramblerCnt--;
			for( int i = 0; i < ScramblerCnt; i++ )
			{
				Scramblers[i] = Scramblers[i + 1];
			}
			Scramblers[ScramblerCnt] = NULL;		
		}
	
		if ( !Island.FindPathBetweenKeypoints( CurrentPoint, final, Path, PathLen ) )
		{
			Print( "WARNING: failed to find path to CC keypoint!" );
		}
		else
		{
			PathPos = 0;

			if ( TestMode )
			{
				//TEST: visualize the fight path
				IslandKeyPoint prev = NULL;
				for( i = 0; i < PathLen; i++ )
				{
					IslandKeyPoint tmp = Path[i];
					tmp.VisualizeAsPath( prev );
					prev = tmp;
				}
			}
		}
	}

	void LaunchNew( IslandModel at, EntityBasic from )
	{
		Island = at;
		MyPlayer = from.GetPlayer();
		Panic = false;
		PostponeOneFrame = true; // one-frame wait for changes commit

		SwitchedToAction = false;
		if ( Island == g_Game.GetCurrentIsland() && MyPlayer != g_Game.GetControlledPlayer() /* HACK: demo mode */ )
			Mode = BATTLE_ACTION;
		else
			Mode = BATTLE_DISTANT;

		if ( MyPlayer == g_Game.GetControlledPlayer() )
			TestMode = true;
		else
			TestMode = false;

		for( int i = 0; i < MAX_UNITS; i++ )
		{
			AssaultUnit unit = Units[i];
			unit.Invalidate();
		}

		Time = g_Game.GetCampaignTime();
		Progress = STATE_ASSAULT;

		CurrentPoint = Island.FindBestLandingSpot( from.GetCenter() );
		FinalPoint = Island.GetFinalCCPoint();
		VisitedCnt = 0;
		FightTimeStamp = -1;
		CurrTeleportIndex = 0;
		PathPos = 0;

		// initialize equipper
		if ( GetCampaign() )
			GetCampaign().InitGenericEquip( MyPlayer );

		// started attack on island
		if ( at.GetPlayer().GetID() != 0 )
		{
			at.OnAttackStart( MyPlayer );

			for( int l = 0; l < g_Game.GetNumIslands(); l++ )
			{
				IslandModel island = g_Game.GetIsland(l);
				if ( island != at )
					island.OnAttackStop( MyPlayer );
			}
		}

		if ( !CurrentPoint )
			Print( "WARNING: no landing keypoint found, the assault won't work properly!" );

		if ( !FinalPoint )
			Print( "WARNING: no final assault keypoint found, the assault won't work properly!" );

		//TEST: first visit all the scramblers
		Island.GetScramblerPoints( Scramblers, ScramblerCnt );
		// remove landing and CC points, will be visited anyway
		for( i = 0; i < ScramblerCnt; i++ )
		{
			IslandKeyPoint kp = Scramblers[i];
			if ( kp == CurrentPoint || kp == FinalPoint )
			{
				Scramblers[i] = NULL;
				ScramblerCnt--;
			}
		}
		CompactPtrArray( Scramblers, 8 );

		if( CurrentPoint && FinalPoint )
		{
			MyPlayer.EnableRepairs( false );
			
#ifdef DEVELOPER
			string mess = "Assault started on island " + at.GetName();
			Print( mess );
#endif
			TowardsNextPoint();
		}

		if ( TestMode )
		{
//			Island.DisablePopulation();
			g_Game.TimeWarp( true ); // speed it up
		}
	}

	// abort current assault
	void Abort( bool winEnd )
	{
		if ( Progress == STATE_DOCK )
			return; // already aborting

		Island.OnAttackStop( MyPlayer );
		if ( !winEnd )
			Panic = true;
			
		// dock all
		for( int i = 0; i < MAX_UNITS; i++ )
		{
			AssaultUnit unit = Units[i];
			if ( unit.AssemblyID < 0 )
				continue;

			EntityBasic assembly = MyPlayer.GetAssembly( unit.AssemblyID );
			if ( Mode == BATTLE_ACTION )
			{
				Unit runit = assembly.GetEntity();
				if ( !assembly.IsDestroyed() && unit.Deployed && runit )
				{
					runit.Dock( REPLACE );
#ifdef DEVELOPER
					widget reason = FindWidget( NULL, "StrategyRoot.MA_KeypointNameValue" );
					SetWidgetText( reason , 0, "Funkce Abort" );
					ShowUnitPlanInDebug( i+2, "DOCK" );
#endif
				}
			}
		}

		HackingCC = false;
		Progress = STATE_DOCK;
		VisitedCnt = 0;

#ifdef DEVELOPER
		widget wdg = FindWidget( NULL, "StrategyRoot.MA_KeypointNameValue" );
		SetWidgetText( wdg, 0, "" );

		wdg = FindWidget( NULL, "StrategyRoot.MA_KeypointTimeValue" );
		SetWidgetText( wdg, 0, "" );

		wdg = FindWidget( NULL, "StrategyRoot.MA_KeypointStatsValue" );
		SetWidgetText( wdg, 0, "" );

		wdg = FindWidget( NULL, "StrategyRoot.MA_PlayerStatsValue" );
		SetWidgetText( wdg, 0, "" );
#endif

	}

	bool PrepareUnits()
	{
		// solves both DISTANT & ACTION
		EntityBasic assemblies[MAX_UNITS];
		int numAssemblies = MyPlayer.GetAssemblies( assemblies );

		if ( MyPlayer.GetCurrentIsland() == g_Game.GetCurrentIsland() && MyPlayer != g_Game.GetControlledPlayer() )
			Mode = BATTLE_ACTION;
		else
			Mode = BATTLE_DISTANT;

		if ( g_Game.IsTimeWarp() )
			Mode = BATTLE_DISTANT;

#ifdef DEVELOPER
		if ( Mode == BATTLE_ACTION )
		{
			widget reason = FindWidget( NULL, "StrategyRoot.MA_KeypointTimeValue" );
			string text = "";
			if ( Progress == 0 )
				text = "NONE"

			if ( Progress == 1 )
				text = "ASSAULT"

			if ( Progress == 2 )
				text = "DOCK"
			
			SetWidgetText( reason , 0, text );
		}
#endif
			
		for( int i = 0; i < numAssemblies; i++ )
		{
			EntityBasic assembly = assemblies[i];
			AssaultUnit unit = Units[i];

			if ( unit.Deployed && assembly.IsDestroyed() )
			{
				// enemy destroyed an unit, validate panic
				unit.Invalidate();

				int total = 0;
				for( int j = 0; j < numAssemblies; j++ )
				{
					unit = Units[j];
					if ( unit.AssemblyID < 0 || !unit.Deployed )
						continue;

					total++;
				}

				if ( total < MinAssaultUnits( MyPlayer, Island ) )
				{
					// lost too much units, back off & flee
					Abort( false );
					Progress = STATE_NONE;

					if ( TestMode )
						g_Game.TimeWarp( false );

					return false;
				}
			}
			else
			{
				if ( unit.AssemblyID < 0 )
				{
					//TODO: check for already deployed unit (can happen after carrier fight for example)
					if ( assembly.IsDeployable() )
					{
				  		Carrier myCarrier = MyPlayer.GetCarrier();
						unit.AssemblyID = MyPlayer.GetAssemblyIndex( assembly );

						int id = assembly.GetSlot();
						// for distant, equip immediately
						//TODO: could equip immediately even for action battle?
						if ( Mode == BATTLE_DISTANT && !TestMode )
						{
							if( id < 5 )
								unit.AutoEquip( true, MyPlayer, 1 );
							else
								unit.AutoEquip( false, MyPlayer, 1 );

							// repair unit before new attack
							assembly.SetHealth( 1.0 );
						}

						// deploy new (only if in action)
						if ( Mode == BATTLE_ACTION && myCarrier )
						{
							if ( assembly.GetEquipmentCount() == 0 )
							{
								if ( id < 5 )
									unit.AutoEquip( true, MyPlayer, 1 );
								else
									unit.AutoEquip( false, MyPlayer, 1 );
							}

							PostponeOneFrame = true; // no need to wait in action mode
							//STUB: need to clear all equip due to some friggin' nonsense in the game code o.O
							//assembly.ClearAllEquipment();
							myCarrier.AssemblyDeploy( assembly.GetSlot() );
						}
					}
				}
				else
				{
					// ready?
					if ( Mode == BATTLE_ACTION )
					{
						if ( !unit.Deployed )
						{
							unit.Deployed = true;
							id = assembly.GetSlot();
							// equip some nice stuff
							if ( !SwitchedToAction )
							{
								if ( id < 5 )
									unit.AutoEquip( true, MyPlayer, 1 );
								else
									unit.AutoEquip( false, MyPlayer, 1 );
							}
							else
							{
								// teleport unit to right location (should be already equipped)
								if ( g_Game.GetCurrentIsland().GetPlayer().GetID() == 0 )
								{
									if ( id == 5 )
									{
										if ( WaitingForTeleport )
											TeleportAfterModeSwitch( unit );
										else
										{
											myCarrier.AssemblyDeploy( unit.AssemblyID );
											unit.RelUnit( MyPlayer).GoTo( g_Game.GetCommandCenter().GetPosition(), REPLACE );
										}
									}
								}
								else
								{
									if ( WaitingForTeleport )
										TeleportAfterModeSwitch( unit );
									else
									{
										myCarrier.AssemblyDeploy( unit.AssemblyID );
										unit.RelUnit( MyPlayer).GoTo( g_Game.GetCommandCenter().GetPosition(), REPLACE );
									}
								}
							}
						}
						else
						{
							//STUB: player has docked some unit
							if ( !assembly.IsActive() )
								unit.Invalidate();
						}
					}
				}
			}
		}

		// all deployed instantly
		if ( Mode == BATTLE_DISTANT )
		{
			WaitingForTeleport = false; // surely not waiting
			return true;
		}

		WaitingForTeleport = false; // surely not waiting
		return true;
	}

	bool PointVisited( IslandKeyPoint point )
	{
		for( int i = 0; i < VisitedCnt; i++ )
		{
			IslandKeyPoint kp = VisitedPoints[i];
			if( kp == point )
				return true;
		}

		return false;
	}

	bool ObtainNextPoint()
	{
		// valid for both ACTION & DISTANT

		if ( !CurrentPoint || !FinalPoint )
			return false;

		// visited CurrentPoint
		if ( !PointVisited( CurrentPoint ) )
		{
			VisitedPoints[VisitedCnt] = CurrentPoint;
			VisitedCnt++;
		}

		// progress path
		PathPos++;
		if ( PathPos < PathLen )
		{
			CurrentPoint = Path[PathPos];
			return true;
		}
		else
		{
			// done with assault
			if ( CurrentPoint == FinalPoint )
				return false;
			else
			{
				TowardsNextPoint();
				return true; // carry on
			}
		}
	}

	// kill some units (TODO: should be based on keypoint/island difficulty)
	// returns true if battle should continue, false if AI should back off
	bool ProcessLoss()
	{
#ifdef DEVELOPER
			Print( "Processing battle results..." );
#endif

		//TEST: split point fight into multiple steps
		int numSteps = 1;//rand( 2, 5 );
		float ptAA = PtAAPower / numSteps;
		float ptAT = PtATPower / numSteps;
		
#ifdef DEVELOPER
		string mess;
		mess = "Num rounds: ";
		mess += itoa( numSteps );
		Print( mess );
#endif
		
		for( int s = 0; s < numSteps; ++s )
		{
			//TEST: DnD style
			float aaUC = PlayerAAPower - PtAADefense;
			float atUC = PlayerATPower - PtATDefense;

			int total = 0;
			float aaDiff;

			if ( PlayerAADefense > 0 )
				aaDiff = ptAA / PlayerAADefense;
			else
				aaDiff = 10.0;

			float atDiff;
			if( PlayerATDefense > 0 )
				atDiff = ptAT / PlayerATDefense;
			else
				atDiff = 10.0;

#ifdef DEVELOPER
			mess = "Starting round ";
			mess += itoa( s + 1 );
			Print( mess );
#endif

			for( int i = 0; i < MAX_UNITS; i++ )
			{
				AssaultUnit unit = Units[i];
				if( unit.AssemblyID < 0 ) continue;

				total++;

				EntityBasic assembly = MyPlayer.GetAssembly( unit.AssemblyID );

				float damage = frand( 0.05, BaseDamage( MyPlayer, Island, assembly ) );

				if ( assembly.IsManta() )
					damage *= aaDiff;
				else
					damage *= atDiff;

				//TEST: healer in our ranks
				if( PlayerHealing > 0 )
					damage *= 0.5;

#ifdef DEVELOPER
				if ( assembly.IsManta() )
				{
					mess = "Manta ";
					mess += itoa( assembly.GetSlot() );
				}
				else
				{
					mess = "Walrus ";
					mess += itoa( assembly.GetSlot() - 4 );
				}
				mess += " took ";
				mess += ftoa( damage );
				mess += " damage";
				Print( mess );
#endif

				float health = assembly.GetHealth() - damage;
				//randomly kill units if the island has higher defense
				if ( GetCampaign().GetIslandForceMod(Island) > 1 )
				{
					int r = 0;
					r = rand(0,5);

					if ( GetCampaign().GetIslandForceMod(Island) > 1.4 )
						r = rand(0,10);

					if ( GetCampaign().GetIslandForceMod(Island) > 1.8 )
						r = rand(0,20);

					if ( r > rand(2,100) )
						health = 0;
				}



				if ( health > 0 )
				{
					assembly.SetHealth( health );
				}
				else
				{
#ifdef DEVELOPER
					Print( "   and was destroyed!" );
#endif
					
					unit.Invalidate();
					assembly.Destroy();
					total--;
				}
			}
		}

		// fight till last man!
		if( total < 1 )
			return false;

		return true;
	}

	void UpdateCombatStats()
	{
		CurrentPoint.CalculateCombatPower( PtAAPower, PtATPower, PtAADefense, PtATDefense );
		PlayerAAPower = 0;
		PlayerATPower = 0;
		PlayerAADefense = 0;
		PlayerATDefense = 0;

		for( int i = 0; i < MAX_UNITS; i++ )
		{
			AssaultUnit unit = Units[i];
			if ( unit.AssemblyID < 0 )
				continue;

			EntityBasic assembly = MyPlayer.GetAssembly( unit.AssemblyID );

			float power;
			float aa;
			float at;
			float aaDef;
			float atDef;
			float support;
			float healing;
			assembly.GetEquipCombatStats( aa, at, aaDef, atDef, support, healing );

			PlayerAAPower += aa;
			PlayerATPower += at;
			PlayerAADefense += aaDef;
			PlayerATDefense += atDef;
			PlayerHealing += healing;

			if ( PlayerAAPower == 0 || PlayerATPower == 0 )
			{
				int index = assembly.GetEquipmentCount();
				int slot1 = -1;
				int slot2 = -1;
				int slot3 = -1;

				if ( index >= 3)
				{
					slot1 = assembly.GetEquipment( 0 );
					slot2 = assembly.GetEquipment( 1 );
					slot3 = assembly.GetEquipment( 2 );
				}
				else
				{
					if ( index == 2 )
					{
						slot1 = assembly.GetEquipment( 0 );
						slot2 = assembly.GetEquipment( 1 );
					}
					else
					{
						if ( index == 1 )
							slot1 = assembly.GetEquipment( 0 );
					}
				}
			}
		}
//This is not needed in strategy mode - we will se how fast the enemy carrier will be now
/*
		float myDmg = g_Game.GetPlayerDamageMultiplier();
		float enDmg = g_Game.GetEnemyDamageMultiplier();

		// multiply by diffivulty settings
		if( MyPlayer == g_Game.GetControlledPlayer() )
		{
			PlayerAAPower *= myDmg;
			PlayerATPower *= myDmg;
			PtAAPower *= enDmg;
			PtATPower *= enDmg;
		}
		else
		{
			PlayerAAPower *= enDmg;
			PlayerATPower *= enDmg;
			PtAAPower *= myDmg;
			PtATPower *= myDmg;
		}
*/
		// calculate capture time modifier
		float diffAA;
		if ( PlayerAAPower > 0 )
			diffAA = PtAADefense / PlayerAAPower;
		else
			diffAA = 10.0;

		float diffAT;
		if ( PlayerATPower > 0 )
			diffAT = PtATDefense / PlayerATPower;
		else
			diffAT = 10.0;

		CapMod = diffAA + diffAT / 2.0;
		if ( CapMod == 0 )
			CapMod = 0.1;
	}

	bool GatherOnPoint()
	{
		if ( !CurrentPoint )
			return true;

		// version for DISTANT:
		if ( Mode == BATTLE_DISTANT )
		{
			float t = g_Game.GetCampaignTime();
			if( FightTimeStamp < 0 )
			{
				// avoid multiple battles for one point
				//TODO: add short delay (still need to travel)
				if ( PointVisited( CurrentPoint ) )
				{
					if ( TestMode )
						CurrentPoint.VisualizeAsCaptured();

					return true;
				}

				UpdateCombatStats();

				FightTimeStamp = t;
				FightDuration = KeypointCapTime( MyPlayer, Island ) * CapMod;

#ifdef DEVELOPER
				string mess = "Battle started over point " + CurrentPoint.GetName();
				Print( mess );

				string name = CurrentPoint.GetName();
				widget stockpileName = FindWidget( NULL, "StrategyRoot.MA_KeypointNameValue" );
				SetWidgetText( stockpileName, 0, name );
				float min = FightDuration / 60;
				min = floor(min);
				name = ftoa( min) + "  min";

				stockpileName = FindWidget( NULL, "StrategyRoot.MA_KeypointTimeValue" );
				SetWidgetText( stockpileName, 0, name );

				mess = "Point stats: AA ";
				mess += ftoa( PtAAPower );
				mess += "/";
				mess += ftoa( PtAADefense );
				mess += " | AT ";
				mess += itoa( PtATPower );
				mess += "/";
				mess += ftoa( PtATDefense );
				Print( mess );
				mess = "Player stats: AA ";
				mess += ftoa( PlayerAAPower );
				mess += "/";
				mess += ftoa( PlayerAADefense );
				mess += " | AT ";
				mess += ftoa( PlayerATPower );
				mess += "/";
				mess += ftoa( PlayerATDefense );
				Print( mess );
				mess = "Fight will take ";
				mess += ftoa( FightDuration );
				Print( mess );

				PtAAPower = floor( PtAAPower );
				PtAADefense = floor( PtAADefense );
				PtATPower = floor( PtATPower );
				PtATDefense = floor( PtATDefense );
				
				PlayerAAPower = floor( PlayerAAPower );
				PlayerAADefense = floor( PlayerAADefense );
				PlayerATPower = floor( PlayerATPower );
				PlayerATDefense = floor( PlayerATDefense );

				name = "AA " + ftoa( PtAAPower ) + " / " +ftoa(PtAADefense) + " | AT " + ftoa(PtATPower) + " / " + ftoa(PtATDefense);
				stockpileName = FindWidget( NULL, "StrategyRoot.MA_KeypointStatsValue" );
				SetWidgetText( stockpileName, 0, name );

				name = "AA " + ftoa( PlayerAAPower ) + " / " +ftoa(PlayerAADefense) + " | AT " + ftoa(PlayerATPower) + " / " + ftoa(PlayerATDefense);
				stockpileName = FindWidget( NULL, "StrategyRoot.MA_PlayerStatsValue" );
				SetWidgetText( stockpileName, 0, name );

				if ( PlayerAAPower == 0 || PlayerATPower == 0 )
				{
					//tohle vypada na spatnej equip, ale proc kurva
					g_Game.ShowHint("DEVELOPER MESSAGE: |Doslo k chybe v equipu jednotek nepratelskeho carrieru!!!", 10, false);
				}

#endif
				if ( FightDuration > MAX_FIGHT_DURATION || Island.IsBuildingCommandCenter() )
				{
					Print("Fight would take too much time, or CC is under construction - aborting...");
					if ( TestMode )
						MyPlayer.EnableRepairs( true );

					Abort( false );

					if ( TestMode )
						g_Game.TimeWarp( false );

					return false;
				}
				
				if ( TestMode )
					CurrentPoint.VisualizeAsFight();
			}
			else
			{
				// finished?
				float dt = t - FightTimeStamp;
				if( dt > FightDuration )
				{
					// calculate loses
					bool lowHP = false;
					if ( !ProcessLoss() )
					{
#ifdef DEVELOPER
						Print( "Lost too much units, aborting..." );
#endif
						if ( TestMode )
							MyPlayer.EnableRepairs( true );

						Abort( false );

						if ( TestMode )
							g_Game.TimeWarp( false );

						return false;
					}
					else
					{
						// point captured
						if ( TestMode )
							CurrentPoint.KillAllUnitsAround();


//TODO: HEAL UNITS
						float totalHP = 0;
						float aliveCnt = 8;
						for( int e = 0; e < 8; e++ )
						{
							EntityBasic assembly = MyPlayer.GetAssembly( e );
							if ( assembly.IsDestroyed() )
							{
								aliveCnt--;
								continue;
							}
	
							totalHP += assembly.GetHealth();
						}
						
						float ratio = totalHP / aliveCnt ;
						float newTime = -1;
						if ( ratio < 0.5 )
						{
							lowHP = true;
							newTime =  ratio * 360;
#ifdef DEVELOPER
							Print( "Wait until units will be healed" );
							Print( newTime );
#endif
						}

					}

					// visualize
					if ( TestMode )
						CurrentPoint.VisualizeAsCaptured();

					if ( lowHP )
					{
						for( int x = 0; x < 8; x++ )
						{
							EntityBasic ass = MyPlayer.GetAssembly( x );
							if ( !ass.IsDestroyed() )
								ass.SetHealth( 1.0 );
						}
						FightDuration = newTime;
					}
					else
						FightTimeStamp = -1;

					return true;
				}
			}

			return false;
		}

		// version for ACTION
		if ( Mode == BATTLE_ACTION )
		{
			// assaultProgress == 1 docking
			// assaultProgress == 2 go to cc
			bool gathered = false;

			CommandCenter cc = g_Game.GetCommandCenter();
			IslandModel island = g_Game.GetCurrentIsland();

			//Capsule check
			bool hasCapsule = false;

			//for capsules check only walruses
			for( int l = 5; l < 9; l++ )
			{
				EntityBasic entBasic = MyPlayer.GetAssembly( l );
				InteractiveEntity entInt = entBasic.GetEntity();

				if ( entBasic.GetEntity() != NULL )
				{
					if ( entInt.GetHealth() > 0 && entBasic.IsFullyDeployed() )		//jedina moznost jak zjistit, ze je jednotka dead, ponevadz assembly neni nikdy null
					{
						int equipmentCount = entBasic.GetEquipmentCount();
						for( int y = 0; y < equipmentCount; y++)
						{
							int equipment = entBasic.GetEquipment( y );
							if ( equipment == capsuleHack || equipment == capsuleDefense || equipment == capsuleProduction || equipment == capsuleResource )
							{
								Progress = STATE_ASSAULT;
								hasCapsule = true;
								break;
							}
						}
					}
				}
			}

			//----------------------------------------------------------------------------
			//we can not take the island if we dont have a capsule
			if ( !hasCapsule && Progress != STATE_DOCK )
			{
				for( int i = 0; i < MAX_UNITS; i++ )
				{
					AssaultUnit unit = Units[i];
					if ( !unit.Deployed )
						continue;

					Unit runit = unit.RelUnit( MyPlayer );
					if ( runit == NULL || runit.IsDocking() )
						continue;
#ifdef DEVELOPER
					ShowUnitPlanInDebug( i+2, "DOCK");
					widget reason = FindWidget( NULL, "StrategyRoot.MA_KeypointNameValue" );
					SetWidgetText( reason , 0, "Neni capsule" );
#endif
					Progress = STATE_DOCK;
					runit.Dock( REPLACE );
				}
				
				return false;
			}
							

			//----------------------------------------------------------------------------
			//Analyze assault state
			if ( island.IsBuildingCommandCenter() || cc.GetPlayer() == MyPlayer )
			{
				//----------------------------------------------------------------------------
				//dock units
				if ( Progress != STATE_DOCK )
				{
					for( i = 0; i < MAX_UNITS; i++ )
					{
						unit = Units[i];
						if ( !unit.Deployed )
							continue;

						runit = unit.RelUnit( MyPlayer );
						if ( runit == NULL || runit.HasPlan() )
							continue;

#ifdef DEVELOPER
						ShowUnitPlanInDebug( i+2, "DOCK");
						reason = FindWidget( NULL, "StrategyRoot.MA_KeypointNameValue" );
						SetWidgetText( reason , 0, "Ostrov je muj, nebo se stavi" );
#endif
						Progress = STATE_DOCK;
						runit.Dock( REPLACE );
					}
				}
			}
			else
			{
				//----------------------------------------------------------------------------
				//send units near CC
				vector CCpos = cc.GetPosition();
				for( i = 0; i < MAX_UNITS; i++ )
				{
					unit = Units[i];
					if ( !unit.Deployed )
						continue;

					runit = unit.RelUnit(MyPlayer);
					if ( runit == NULL )
						continue;

					vector unitPos = runit.GetPosition();
					vector to = CCpos - unitPos;
					to[2] = 0; // flat distance
					float dist = VectorNormalize( to );
					if ( dist < GATHER_DIST_SQ )
					{
						if ( runit.HasPlan() && !runit.HasEquip( capsuleHack ) )
							runit.PlanClear();

						if ( runit.HasEquip( capsuleHack ) )
						{	
							if ( !HackingCC )
							{
								runit.PlanClear();
								HackingCC = true;
								bool test = runit.InvokeHackOfBuilding( cc );
								CCTimeStamp = g_Game.GetCampaignTime();
								cc.HackByPlayer( MyPlayer );
								gathered = true;
								break;
							}
							else
							{
								if ( island.GetPlayer() != MyPlayer )
								{
									float hackingTime = CCTimeStamp + MAX_CC_HACKING_TIME;
									//float time = g_Game.GetCampaignTime();
									if ( g_Game.GetCampaignTime() > hackingTime )
									{
										HackingCC = false;
										cc.HackByPlayer( MyPlayer );
										gathered = true;
										break;
									}
								}
							}
						}
						else
						{
							if ( runit.HasEquip( capsuleDefense ) )
							{
								Island.BuildCommandCenter( MyPlayer, capsuleDefense );
								gathered = true;	
							}

							if ( runit.HasEquip( capsuleProduction ) )
							{
								Island.BuildCommandCenter( MyPlayer, capsuleProduction );
								gathered = true;	
							}

							if ( runit.HasEquip( capsuleResource ) )
							{
								Island.BuildCommandCenter( MyPlayer, capsuleResource );
								gathered = true;	
							}								
						}
					}
					else
					{
						gathered = false;
						break;
					}
				}
			}
		}

		if ( gathered )
		{
			if ( Progress == STATE_ASSAULT )
			{
				MyPlayer.DockAllUnits();
				Abort( true );
#ifdef DEVELOPER
				ShowUnitPlanInDebug( 0, "DOCK");
				ShowUnitPlanInDebug( 1, "DOCK");
				ShowUnitPlanInDebug( 2, "DOCK");
				ShowUnitPlanInDebug( 3, "DOCK");

				ShowUnitPlanInDebug( 4, "DOCK");
				ShowUnitPlanInDebug( 5, "DOCK");
				ShowUnitPlanInDebug( 6, "DOCK");
				ShowUnitPlanInDebug( 7, "DOCK");

				reason = FindWidget( NULL, "StrategyRoot.MA_KeypointNameValue" );
				SetWidgetText( reason , 0, "Seskupili jsme se" );
#endif
			}
		}
		else
		{
			if ( cc && hasCapsule )
			{
				if ( !island.IsBuildingCommandCenter() && cc.GetPlayer() != MyPlayer )
				{
					for( i = 0; i < MAX_UNITS; i++ )
					{
						unit = Units[i];
						if ( unit.AssemblyID < 0 )
							continue;

						assembly = MyPlayer.GetAssembly( unit.AssemblyID );
						runit = assembly.GetEntity();
						if ( !assembly.IsDestroyed() && unit.Deployed && runit )
						{
							if ( runit.HasPlan() && runit.IsDocking() )
								runit.ClearPlan();

							if ( !runit.HasPlan() )
							{
								runit.GoTo( CCpos, REPLACE );
#ifdef DEVELOPER
								ShowUnitPlanInDebug( i+2, "CC");
								reason = FindWidget( NULL, "StrategyRoot.MA_KeypointNameValue" );
								SetWidgetText( reason , 0, "Jedem znova k CC" );
#endif
							}
						}
					}
				}
			}
		}

		return gathered;
	}

	bool Conquer()
	{
		// valid for both ACTION & DISTANT
		if ( TestMode )
		{
			Print( "You have won the battle!" );
			MyPlayer.EnableRepairs( true );
		}
		else
		{
			Print( "Island was captured!" );
		}

		//TODO: NO API FOR BUILDING CommandCenter BY AI
		if ( Island.GetPlayer().GetID() == 0 && !Island.IsBuildingCommandCenter() )
		{
			GameStorage carrStock = MyPlayer.GetCarrierStorage();

			if ( g_Game.MAX_SINGLE_PRODUCTION == 0 )
				g_Game.GetGameIslandsIncomeSettings();
			
			int type = g_Game.GetBestIslandTypeScript( MyPlayer, Island.GetArea() )
			switch( type )
			{
				//3 == Factory
				case 3 :
					Island.BuildCommandCenter( MyPlayer, capsuleProduction );
					break;
				
				//4 == Defence
				case 4 :
					Island.BuildCommandCenter( MyPlayer, capsuleDefense );
					break;
				
				//5 == Resource
				case 5:
					Island.BuildCommandCenter( MyPlayer, capsuleResource );
					break;
					
				default:
					Island.BuildCommandCenter( MyPlayer, capsuleDefense );
					break;
			}
		}
		else
		{
			Island.AssignToPlayer( MyPlayer, true );
			if ( Island.GetType() == IslandNeutral )
				Island.SetType( IslandDefence );
		}

		if ( TestMode )
		{
			MyPlayer.EnableRepairs( true );
			g_Game.TimeWarp( false );
		}

		// stopped attack on island
		Island.OnAttackStop( MyPlayer );
		Abort( true );
		return true;
	}

	void Update()
	{
#ifdef DEVELOPER
		if ( KeyState( KC_RSHIFT ) > 0 && MyPlayer != g_Game.GetControlledPlayer() )
		{
			debug
			//Assault script
			ClearKey( KC_RSHIFT );
		}
#endif

		if ( Progress == STATE_NONE )
			return;

		if ( PrepareUnits() )
		{
			//REASONING: we need to postpone combat logic because the equipment we assigned to our
			// assemblies will take one frame to process (we would be conquering with zero strength)
			if ( PostponeOneFrame )
			{
				PostponeOneFrame = false;
				return;
			}

			// do not update yet
			if ( !WaitingForTeleport )
			{
				if ( GatherOnPoint() )
				{
					Print( CurrentPoint );
					if ( CurrentPoint == FinalPoint )
						Print( "At final point!" );

					if ( !ObtainNextPoint() )
					{
						Conquer();
						Progress = STATE_NONE;
					}
				}
			}
		}
	}

	void SwitchMode()
	{
	if ( !MyPlayer )
		return;
		if ( Mode == BATTLE_ACTION )
		{
			// switch to distant
			Mode = BATTLE_DISTANT;

			// entities will be removed automatically, all we need to do is continue from
			// current point as usual (+ should fight for current point for a limited while?)

			// clear entity pointers
			MyPlayer.InstatDockAllUnits();
/*
			for( int i = 0; i < MAX_UNITS; i++ )
			{
				AssaultUnit unit = Units[i];
				EntityBasic entBasic = unit.BasicUnit( MyPlayer );
				
				if ( entBasic != NULL && entBasic.IsFullyDeployed() )
					entBasic.InstantDock();
				
				unit.Deployed = false;
			}
*/
			SwitchedToAction = false;
		}
		else
		{
			// switch to action
			Mode = BATTLE_ACTION;

			//algorithm: kill enemy units on visited points, fill the entity array, teleport alive
			// units to current point and continue as usual
			CurrentPoint.KillAllUnitsAround();
			ProcessActionSwitch();

			SwitchedToAction = true;
		}
	}
};
/**@}*/